/* File: sdl-hjm.c */

/*
 * Copyright (c) 1997 Ben Harrison
 * Portions Copyright (c) 2001 Gregory Velichansky (details in main-sdl.c)
 *
 * This software may be copied and distributed for educational, research,
 * and not for profit purposes provided that this copyright and statement
 * are included in all such copies.
 *
 * This code may also be used under the GNU GPL. 
 *
 * If you do not read msin-sdl.c, you have NO RIGHT to distribute this code, 
 * use it for any purpose, distribute in any form or for any reason, or store
 * it in any fashion and must destroy all copies of this code.
 *
 */


#include "angband.h"
#include "main-sdl.h"

#ifdef USE_HJM
/*
 * Simugraph system wrapper header (Hj. Malthaner)
 */
#include "simugraph/simsys.h"


/*
 * Iso view world adaptor (Hj. Malthaner)
 */
#include "hajo_src/world_adaptor.h"


/*
 * Text place marker function protype (Hj. Malthaner)
 */
static void set_spots(int x, int y, int n, bool v);



typedef struct graf_data graf_data;

struct graf_data
{
	int happy; /* are we having fun yet? */

	/* Your special state information goes here.
	 * ie:
	SDL_Surface *tiles;

	   or perhaps, if your tiles are simple bitmaps, you can take advantage of
	
	font_data *tiles;
	 */
};


static graf_data graphics; /* change this line. */

static term_data *data; /* to be initialized in init_hjm() */

/*
 * Simugraph specific routines
 * by Hj. Malthaner
 */
#include "hajo_src/hackdef.h"


/**
 * we need to track spots with text to avoid overdrawing text with images
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
bool spots[80][24]; /* this is hard-coded rubbish. use halloc() here.  FIXME */


/**
 * spot array access prozedure
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
static void set_spots(const int x, const int y, const int n, const bool v)
{
		int i;

		for(i=x; i<x+n; i++) {
			spots[i][y] = v;
		}
} 



unsigned char **iso_ap;
unsigned char **iso_cp;
unsigned char **iso_atp;
unsigned char **iso_ctp;




/*
 * Hajo: this flags need to be set when opening windows
 */

static int tex_width;
static int tex_height;
static int tex_xoff;
static int tex_yoff;


static unsigned int tab16[1<<16];

/** 
 * mouse coordinates for Simugraph engine 
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */    
int mx, my; 


/**
 * inits operating system stuff
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
int dr_os_init(int n, int *parameter)
{
    // Hajo:
    // unused in isov-x11
    return TRUE;
}


/**
 * opens graphics device/context/window of size w*h
 * @param w width
 * @param h height
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
int dr_os_open(int w, int h)
{
    const int left = 13;

    tex_width  = (data[0].w * (data[0].t.wid-left+2) + 3) & 0xFFFC;
    tex_height = data[0].h * (data[0].t.hgt-2);
                                                  
    tex_xoff = data[0].w * left;
    tex_yoff = data[0].h * 1 + 1;

    return TRUE;
}


/**
 * closes operating system stuff
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
int dr_os_close()
{
    // Hajo:
    // unused in isov-x11
    return TRUE;
}


/**
 * retrieve display width
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
int dr_get_width()
{
    return tex_width;
}


/**
 * retrieve display height
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
int dr_get_height()
{
    return tex_height;
}


/**
 * this is used if we need to fake an 8 bit array
 * @author Hj. Malthaner
 */
static unsigned short * data8;


/**
 * creates a (maybe virtual) array of graphics data
 * @author Hj. Malthaner
 */
unsigned short * dr_textur_init()
{       
    int i;

    printf("isov-x11::dr_textur_init()\n");
    printf(" width  = %d\n", tex_width);
    printf(" height = %d\n", tex_height);

    for(i=0; i<(1<<16); i++) {
	// FIXME!!!
	// must consider color bits, or breaks in anything else but RGB 555
	unsigned int R;
	unsigned int G;
	unsigned int B;
                       
	// RGB 555
	R = (i & 0x7C00) >> 10;
	G = (i & 0x03E0) >> 5;
	B = (i & 0x001F) >> 0;

	 
	tab16[i]= SDL_MapRGB(data[0].face->format, R<<3, G<<3, B<<3);
    }         


            
    data8 = malloc((tex_width)*(tex_height)*2);

    printf(" textur = %p\n", data8);

    // fake an 16 bit array and convert data before displaying
    return data8;
}


static void flush_area(int dest_x, int dest_y,
                       int x, int y, int w, int h)
{
    SDL_Surface *face = data[0].face;

    if(SDL_LockSurface(face) == 0) {
	int i, j;
	const int bpp = face->format->BytesPerPixel;
      
	for(j=0; j<h; j++) {
	    unsigned short * p  = data8 + (y+j)*tex_width + x;
	    unsigned char * row = face->pixels + (dest_y+j) * face->pitch + dest_x * bpp; 


	    for(i=0; i<w; i++) {
		*((unsigned short*)row) = tab16[ *p++ ];
		row += bpp;
	    }
	} 
	SDL_UnlockSurface(face);
    }
}

/**
 * displays the array of graphics data
 * @author Hj. Malthaner
 */
void dr_textur(int xp, int yp, int w, int h)
{
    int x,y;

    // clipping unten
    if(yp+h > tex_height) {
	h = tex_height-yp;
    } 

    /* debug spots
    for(y=0; y<24; y++) {
	for(x=0; x<80; x++) {
	    if(spots[x][y]) {
		printf("X");
	    } else {
		printf(".");
	    }
	}
	printf("\n");
    }
    */

    for(y=0; y<SCREEN_HGT; y++) {
	const int left = 13;
	const int y1 = y+1;
	int x = 0;

	yp = data[0].h * y;

	spots[79][y1] = FALSE;

	do {
	    int n = 0;
	    while(x+n+left < 80 && !spots[x+n+left][y1]) {
		n++;
	    } 

	    xp = data[0].w * x;


	    flush_area(tex_xoff+xp, tex_yoff+yp,
                       xp, yp, 
		       data[0].w*(n), data[0].h);

	    x += n;

	    while(x+left < 80 && spots[x+left][y1]) {
		x++;
	    } 	    
	} while(x+left < 80);      
    }
}


/**
 * use this method to flush graphics pipeline (undrawn stuff) onscreen.
 * @author Hj. Malthaner
 */
void dr_flush()
{
    // Iso-view for angband needs no sync.
    // XSync(md,FALSE);
}


/**
 * set colormap entries
 * @author Hj. Malthaner
 */ 
void dr_setRGB8multi(int first, int count, unsigned char * data)
{
    // Hajo:
    // unused in isov-x11    
}


/**
 * display/hide mouse pointer
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
void show_pointer(int yesno)
{
    // Hajo:
    // unused in isov-x11
}


/**
 * move mouse pointer
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
void move_pointer(int x, int y)
{
    // Hajo:
    // unused in isov-x11
}


/**
 * update softpointer position
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
void ex_ord_update_mx_my()
{
    // Hajo:
    // unused in isov-x11
}


/**
 * get events from the system
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
void GetEvents()
{
    // Hajo:
    // unused in isov-x11
}


/**
 * get events from the system without waiting
 * @author Hj. Malthaner (hansjoerg.malthaner@gmx.de)
 */
void GetEventsNoWait()
{
    // Hajo:
    // unused in isov-x11
}


/**
 * @returns time since progrma start in milliseconds
 * @author Hj. Malthaner
 */
long long dr_time(void)
{
    // Hajo:
    // unused in isov-x11
    return 0;
}


/**
 * sleeps some microseconds
 * @author Hj. Malthaner
 */
void dr_sleep(unsigned long usec)
{
    // Hajo:
    // unused in isov-x11
}


/**
 * loads a sample
 * @return a handle for that sample or -1 on failure
 * @author Hj. Malthaner
 */
int dr_load_sample(const char *filename)
{
    // Hajo:
    // unused in isov-x11
    return TRUE;
}


/**
 * plays a sample
 * @param key the key for the sample to be played
 * @author Hj. Malthaner
 */
void dr_play_sample(int key, int volume) 
{
    // Hajo:
    // unused in isov-x11
}

/* Large chunk of code copied straight from Hajo's file ends here XXX XXX XXX */





static void Term_init_hjm(term *t)
{
	term_data *td = (term_data*)(t->data);
	graf_data *gd = (graf_data*)(td->graf);

	/* your code here. $5 */

	td->init_hook(t);

}



static void Term_nuke_hjm(term *t)
{
	term_data *td = (term_data*)(t->data);
	graf_data *gd = (graf_data*)(td->graf);

	/* Nuke your stuff first. */

	td->nuke_hook(t);

}

static errr Term_user_hjm(int n)
{
	term_data *td = (term_data*)(Term->data);
	graf_data *gd = (graf_data*)(td->graf);

	/* If you think of a use for this, good for you. */


	return td->user_hook(n);
}


static errr Term_xtra_hjm(int n, int v)
{
	term_data *td = (term_data*)(Term->data);
	graf_data *gd = (graf_data*)(td->graf);

	/* Analyze */
	switch (n)
	{
		/* You probably don't need to process events. (Unless you want the mouse?)
		case TERM_XTRA_EVENT:
		{
			return (0);
		}

		case TERM_XTRA_FLUSH:
		{

			return (0);
		}
		*/

		/* Unless you need to do something special, and you may, let main-sdl.c do it. */
		case TERM_XTRA_CLEAR:
		{
			break;
		}
		

		/* I suggest you implement this one: */
		case TERM_XTRA_SHAPE:
		{
			/*
			 * Set the cursor visibility XXX XXX XXX
			 *
			 * This action should change the visibility of the cursor,
			 * if possible, to the requested value (0=off, 1=on)
			 *
			 * This action is optional, but can improve both the
			 * efficiency (and attractiveness) of the program.
			 */

			/* Handle this yourself if you implement a cursor (which you should) */

			return (0);


		}

		/* This event probably doesn't fit into non-top-down graphics models
		case TERM_XTRA_FROSH:
		{
			return (0);
		}
		*/

		/* You probably do not need to do anything for FRESH, actually.
		 * If you cannot use FROSH, set td->prefer_fresh to ask for
		 * full-time SDL_Flip() on TERM_XTRA_FRESH in main-sdl.c
		 */
		case TERM_XTRA_FRESH:
		{
			if (td->t.higher_pict)refresh_display();

			break;
		}

		/*
		case TERM_XTRA_NOISE:
		{
			return (0);
		}


		case TERM_XTRA_SOUND:
		{

			return (0);
		}*/

		case TERM_XTRA_BORED:
		{
			/* Please, feel free to do something here... */

			break;
		}

		/*  Hopefully, you don't need these either.
		case TERM_XTRA_REACT:
		{
			return (0);
		}
		

		case TERM_XTRA_ALIVE:
		{

			return (0);
		}

		case TERM_XTRA_LEVEL:
		{

			return (0);
		}

		case TERM_XTRA_DELAY:
		{

			return (0);
		}
		*/

		default:
			break;
	}

	/* Unknown or Unhandled action */
	return td->xtra_hook(n, v);
}


static errr Term_curs_hjm(int x, int y)
{
	term_data *td = (term_data*)(Term->data);
	graf_data *gd = (graf_data*)(td->graf);

	/* XXX XXX XXX */

	/* Success */
	return (0);
}


static errr Term_wipe_hjm(int x, int y, int n)
{
	term_data *td = (term_data*)(Term->data);
	graf_data *gd = (graf_data*)(td->graf);

	set_spots(x, y, 1, FALSE);

	return td->wipe_hook(x, y, n);
}


/* Probably you should leave the text drawing to main-sdl.c */
static errr Term_text_hjm(int x, int y, int n, byte a, const char *cp)
{
	term_data *td = (term_data*)(Term->data);

	set_spots(x, y, n, (a < 16));


	return td->text_hook(x, y, n, a, cp);
}


/*
 * Draw some attr/char pairs on the screen
 *
 * This routine should display the given "n" attr/char pairs at
 * the given location (x,y).  This function is only used if one
 * of the flags "always_pict" or "higher_pict" is defined.
 *
 * You must be sure that the attr/char pairs, when displayed, will
 * erase anything (including any visual cursor) that used to be at
 * the given location.  On many machines this is automatic, but on
 * others, you must first call "Term_wipe_hjm(x, y, 1)".
 *
 * With the "higher_pict" flag, this function can be used to allow
 * the display of "pseudo-graphic" pictures, for example, by using
 * the attr/char pair as an encoded index into a pixmap of special
 * "pictures".
 *
 * With the "always_pict" flag, this function can be used to force
 * every attr/char pair to be drawn by this function, which can be
 * very useful if this file can optimize its own display calls.
 *
 * This function is often associated with the "arg_graphics" flag.
 *
 * This function is only used if one of the "higher_pict" and/or
 * "always_pict" flags are set.
 */
static errr Term_pict_hjm(int x, int y, int n, const byte *ap, const char *cp)
{
	term_data *td = (term_data*)(Term->data);
	graf_data *gd = (graf_data*)(td->graf);

	/*
	 * This code wasn't in Hajo's file and shouldn't be used at all.
	 * It's just here to make me happy. -hmaon
	 */
	memcpy(&iso_ap[y][x], ap, n);
	memcpy(&iso_cp[y][x], cp, n);
	memcpy(&iso_atp[y][x], ap, n);
	memcpy(&iso_ctp[y][x], cp, n);
	
	set_spots(x, y, n, FALSE);


#ifndef USE_TRANSPARENCY
	return td->pict_hook(x, y, n, ap, cp); /* for testing (or if you wish to allow
															toggling of graphics types?) */
#else
	/* this function shouldn't be called if transparency is compiled in.
	 * let's just do something that will compile without warnings. 
	 * hey, it probably even does the right thing.
	 */
	return td->pict_hook(x, y, n, ap, cp, ap, cp);
#endif
}

Term_pict_hjm_28x(int x, int y, const byte a, const char c)
{
	Term_pict_hjm(x, y, 1, &a, &c);
}

/*
 * Hajo:
 * This hook is used by the iso-view
 */
#ifdef USE_TRANSPARENCY
static errr Term_pict_hjm_trans(int x, int y, int n, const byte *ap, const char *cp, const byte *tap, const char *tcp)
{
	term_data *td = (term_data*)(Term->data);
	graf_data *gd = (graf_data*)(td->graf);

#ifndef USE_SMALL_ISO_HACK

	memcpy(&iso_ap[y][x], ap, n);
	memcpy(&iso_cp[y][x], cp, n);
	memcpy(&iso_atp[y][x], tap, n);
	memcpy(&iso_ctp[y][x], tcp, n);
#endif

	set_spots(x, y, n, FALSE);


	Term_pict_hjm(x, y, n, ap, cp);
}

Term_pict_hjm_trans_28x(int x, int y, int n, const byte a, const char c, const byte ta, const char tc)
{
	Term_pict_hjm_trans(x, y, 1, &a, &c, &ta, &tc);
}
#endif

/*** Internal Functions ***/


/*
 * Instantiate a "term_data" structure
 *
 * This is one way to prepare the "term_data" structures and to
 * "link" the various informational pieces together.
 *
 * This function assumes that every window should be 80x24 in size
 * (the standard size) and should be able to queue 256 characters.
 * Technically, only the "main screen window" needs to queue any
 * characters, but this method is simple.  One way to allow some
 * variation is to add fields to the "term_data" structure listing
 * parameters for that window, initialize them in the "init_hjm()"
 * function, and then use them in the code below.
 *
 * Note that "activation" calls the "Term_init_hjm()" hook for
 * the "term" structure, if needed.
 *
 * This function is called at the end of main-sdl.c's term_data_link
 * before "activation." You should use it to overwrite main-sdl.c's default
 * values. If, and only if, this function gets called, the user wants to 
 * use this graphics engine. 
 *
 * The most impotant thing here is hook initialization. It is appropriate
 * to call term_init() here since it will not be called from main-sdl.c's
 * term_data_link() if an alternate graphics engine is requested.
 *
 * Likely default flags have been uncommented. Fix them if they're not right
 * for you.
 */
static void term_data_link(term_data *data, int i)
{
	term_data *td = &data[i];
	term *t = &td->t;

	graf_data *gd = td->graf = &graphics;

	if (i == 0) td->prefer_fresh = TRUE;

	td->cursor_on = FALSE; /* until Hajo implements a cursor with Simugraph */

	/* Initialize the term */
	term_init(t, 80, 24, 256);

	/* Choose "soft" or "hard" cursor XXX XXX XXX */
	/* A "soft" cursor must be explicitly "drawn" by the program */
	/* while a "hard" cursor has some "physical" existance and is */
	/* moved whenever text is drawn on the screen.  See "z-term.c". */
	t->soft_cursor = TRUE;

	/* Avoid the "corner" of the window XXX XXX XXX */
	/* t->icky_corner = TRUE; */

	/* Use "Term_pict()" for all attr/char pairs XXX XXX XXX */
	/* See the "Term_pict_hjm()" function above. */
	/* t->always_pict = TRUE; */

	/* Use "Term_pict()" for some attr/char pairs XXX XXX XXX */
	/* See the "Term_pict_hjm()" function above. */
	if (i == 0) t->higher_pict = TRUE;

	/* Use "Term_text()" even for "black" text XXX XXX XXX */
	/* See the "Term_text_hjm()" function above. */
	/* t->always_text = TRUE; */

	/* Ignore the "TERM_XTRA_BORED" action XXX XXX XXX */
	/* This may make things slightly more efficient. */
	t->never_bored = TRUE;

	/* Ignore the "TERM_XTRA_FROSH" action XXX XXX XXX */
	/* This may make things slightly more efficient. */
	/*t->never_frosh = TRUE; */

	/* Erase with "white space" XXX XXX XXX */
	t->attr_blank = TERM_WHITE;
	t->char_blank = ' ';

	/* Prepare the init/nuke hooks */
	t->init_hook = Term_init_hjm;
	t->nuke_hook = Term_nuke_hjm;

	/* Prepare the template hooks */
	t->user_hook = Term_user_hjm;
	t->xtra_hook = Term_xtra_hjm;
	t->curs_hook = Term_curs_hjm;
	t->wipe_hook = Term_wipe_hjm;

#ifndef OLD_TERM_28X
#ifdef USE_TRANSPARENCY
	t->pict_hook = Term_pict_hjm_trans;
#else
	t->pict_hook = Term_pict_hjm;
#endif
#else /* OLD_TERM_28X */
#ifdef USE_TRANSPARENCY
	t->pict_hook = Term_pict_hjm_trans_28x;
#else
	t->pict_hook = Term_pict_hjm_28x;
#endif
#endif /* OLD_TERM_28X */

	/* let main-sdl.c handle text, probably... */
	t->text_hook = Term_text_hjm;


	/* Remember where we came from */
	t->data = (vptr)(td);

	/* If you expect special screen dimensions, speak now 
	 * or forever hold your peace. XXX */
#if 0
	td->width = 640;
	td->height = 480;
#endif

	/* Please don't mess with td->bpp unless you have a very
	 * good reason to, though. It's best to set it in defaults_sdl_hjm() */


	/* Activate it */
	Term_activate(t);

	if (i == 0)
	{
		init_adaptor();

		printf("Term size is %dx%d\n", data[0].t.wid, data[0].t.hgt);
		iso_cp = halloc(data[0].t.wid, data[0].t.hgt);
		iso_ap = halloc(data[0].t.wid, data[0].t.hgt);
		iso_ctp = halloc(data[0].t.wid, data[0].t.hgt);
		iso_atp = halloc(data[0].t.wid, data[0].t.hgt);
	}
	/* Global pointer */
	/*angband_term[i] = t;*/ /* this doesn't really do anything... blah... XXX */
}

/*
 * This function is called in main-sdl.c to set proper default values for
 * this module. You may expand the list of arguments if you wish.
 * This is an awkward looking trick to isolate the graphics engine module
 * from main-sdl.c as much as possible to facilitate parallel development.
 */
/* fontname is a 64 char-long array */
void defaults_sdl_hjm(char *fontname, Uint8 *bpp)
{
	strcpy(fontname, "font/12x24-blah-jis.hex");
	*bpp = 16;
}



static unsigned char ** halloc(int w, int h)
{
	unsigned char **field = (unsigned char **)malloc(sizeof(unsigned char *) * h);
	int i;

	for(i=0; i<h; i++) {
		field[i] = (unsigned char *)malloc(sizeof(unsigned char) * w);
		memset(field[i], 32 , w);
	}

	return field;
}



/*
 * Initialization function
 * This is called from init_sdl() if this graphics engine is to be used.
 *
 * You may want command line parameters. 
 */
typedef void (*link_callback)(term_data *, int);

link_callback init_hjm(int argc, char **argv, term_data *_data)
{


	/* Do you need to initalize something really special? Do you need to load
	 * your graphics? Do you need to convert the default Angband tiles to your
	 * own format? (Oh yes you do!) This is a good place to do it.
	 */

	data = _data;

	ANGBAND_GRAF = "hjm";
	


	/* Success */
	return term_data_link; /* On error, please return NULL */
}




#endif /* USE_XXX */
